#include "solution.h"


solutionhs *find_solution(solutionhs *P,int pid) {

    solutionhs *s;

    HASH_FIND_INT(P, &pid, s);

    return s;

}

void add_solution(solutionhs **P,int pid, int m){

    solutionhs *s;

    s = calloc(1, sizeof(solutionhs));

    s->pid = pid;
    s->m = m;

    //printf("\n\npid:%d mes gerado:%d",pid,m);

    HASH_ADD_INT(*P, pid, s);

}


void add_fitness_to_solution(solutionhs **P,float fitness){

    solutionhs *s;

    s = calloc(1, sizeof(solutionhs));

    s->pid = -1;
    s->m = 61;
    s->fitness = fitness;

    HASH_ADD_INT(*P, pid, s);

}

void add_population_classification(solutionhs **P,int classification){

    solutionhs *s;

    s = find_solution(*P,-1);

    s->subpopulation = classification;

}


void print_solutions(solutionhs **P){

    solutionhs *s;

    for(s = *P; s != NULL; s =(solutionhs*) (s->hh.next)){
        if(s->pid == -1){
            printf("\nP id %d : m = %d : fitness = %.2f",s->pid,s->m,s->fitness);
        }else{
            printf("\nP id %d : m = %d",s->pid,s->m);
        }

    }
}



void delete_from_solution(solutionhs **P,int pid) {
    solutionhs *r;
    HASH_FIND_INT(*P, &pid, r);
    HASH_DEL(*P, r);
    free(r);
}



void schedule_mandatory_projects_in_solution(solutionhs **P,projectsh **mandatory_projects){

    projectsh *s,*l;

    HASH_ITER(hh, *mandatory_projects, s, l){

        add_solution(P,s->pid,s->project.initialMonth);

    }
}


void copy_solution(solutionhs **P,solutionhs **Q){

    solutionhs *s,*l;
    HASH_ITER(hh,*P,s,l){

        if(s->pid != -1){
            add_solution(Q,s->pid,s->m);
        }else{
            add_fitness_to_solution(Q,s->fitness);
        }
    }
}

void copy_solution_without_fitness(solutionhs **P,solutionhs **Q){

    solutionhs *s,*l;
    HASH_ITER(hh,*P,s,l){

        if(s->pid != -1){
            add_solution(Q,s->pid,s->m);
        }
    }
}
float calcule_all_wp_risk(warning_pointshs **wp){

    warning_pointshs *s,*l;
    float fitness = 0;

    HASH_ITER(hh,*wp,s,l){

        fitness = fitness + s->warning_points.risk;

    }

    return fitness;
}


int is_wp_controlled(solutionhs **Q,warning_pointshs **wp,projectsh **projects,int wpid){


    warning_pointshs *s;
    solutionhs *q;
    projectsh *t;

    /*Encontrar o ponto de ateno correspondente na hash de pontos de ateno*/
    s = find_warning_point(*wp,wpid);

    /*Percorrer o vetor de projetos desse ponto de ateno*/
    for(int i = 0; i< s->warning_points.tam;i++){

        /*Encontrar o projeto na soluo para saber seu mes de agendamento*/
        if(!(q = find_solution(*Q,s->warning_points.group_of_projects[i]))){
            //printf("\n\nNao encontrado na solucao");
            return FALSE;
        }
        /*Encontrar o projeto na hash de soluo para saber seu temmpo de durao*/
        t = find_project(*projects,s->warning_points.group_of_projects[i]);

        /*Verifica se ms de termino(inicio  + durao -1 ) do projeto for maior que 60 ento retoran falso*/
        //printf("\n\nwpid: %d pid: %d q->m:%d t->project.duration:%d",wpid,s->warning_points.group_of_projects[i],q->m,t->project.duration);
        if(q->m + t->project.duration -1 > T){
            //printf("\n\nwpid:%d projeto:%d %d + %d Ultrapassou T:%d.",s->pid,t->pid,q->m,t->project.duration,T);
            return FALSE;
        }
    }

    return TRUE;

}


int calcule_fw(solutionhs **Q,warning_pointshs **wp,projectsh **projects,int wpid){

    warning_pointshs *s;
    solutionhs *q;
    projectsh *t;
    int max = 0;

    /*Encontrar o ponto de ateno correspondente na hash de pontos de ateno*/
    s = find_warning_point(*wp,wpid);

    /*Percorrer o vetor de projetos desse ponto de ateno*/
    for(int i = 0; i< s->warning_points.tam;i++){

        /*Encontrar o projeto na soluo para saber seu mes de agendamento*/
        q = find_solution(*Q,s->warning_points.group_of_projects[i]);
        /*Encontrar o projeto na hash de soluo para saber seu temmpo de durao*/
        t = find_project(*projects,s->warning_points.group_of_projects[i]);

        /*Se o mes agendado mais a durao for maior que max, max recebe esse valor
         max retornar o ms mais tardio de termino desse que esse ponto de ateno  controlado
        */
        if(q->m + t->project.duration - 1 > max){
            max = q->m + t->project.duration -1;
        }
    }

   // printf("\n\nmax:%d",max);
    return max;
}
int sum_wp_controlled_by_solution(solutionhs **Q,warning_pointshs **wp,projectsh **projects){

    warning_pointshs *s,*l;
    float controlled_risk = 0;
    float aux=0;

    /*Percorre toda a hash de pontos de ateno*/
    HASH_ITER(hh,*wp,s,l){

        /*Verifica se o ponto de ateno  controlado.*/
        if(is_wp_controlled(Q,wp,projects,s->pid)){
            /*Se for controlado pela soluo o grau de risco  somado*/

            //printf("\n\nid:%d",s->pid);

            //printf("\n\nfw:%d",calcule_fw(Q,wp,projects,s->pid));

            aux = 2*T - calcule_fw(Q,wp,projects,s->pid);

            controlled_risk += s->warning_points.risk * aux;


        }
    }
   // printf("\n\ncontrolled_risk:%.2f",controlled_risk);
    return controlled_risk;

}


float fitness(solutionhs **Q,projectsh **projects,warning_pointshs **wp){


    return 2 * T * calcule_all_wp_risk(wp) - sum_wp_controlled_by_solution(Q,wp,projects);

}


void swap_solution(solutionhs **A,solutionhs **B){

    solutionhs *aux;

    aux = *A;
    *A = *B;
    *B = aux;

}


void insertion_sort(solutionhs *vetor[],int tam){              /*Ordenao por insero*/

    int i, j;
    solutionhs *atual = NULL;
    solutionhs *s,*q;

    for (i = 1; i < tam; i++)
    {
        atual = vetor[i];
        j = i - 1;

        s = find_solution(atual,-1);
        q = find_solution(vetor[j],-1);

        //printf("\ns->rp: %.2f",s->rp);
        //printf("\nq->rp: %.2f",q->rp);

        while ((j >= 0) && (s->fitness < q->fitness)){

            swap_solution(&vetor[j+1],&vetor[j]);
            j = j - 1;

            if(j != -1){
                q = find_solution(vetor[j],-1);
            }

        }

        vetor[j + 1] = atual;
    }

}

void schedule_projects_that_are_not_in_solution(projectsh **projects,solutionhs **P){

    projectsh *s,*l;
    solutionhs *q;

    HASH_ITER(hh,*projects,s,l){

        q = find_solution(*P,s->pid);

        /*Se o projeto no est na soluo.*/
        if(!(q)){
            add_solution(P,s->pid,61);
        }
    }
}

void generate_solution(projectsh **projects, projectsh **mandatory_projects,warning_pointshs **wp ,solutionhs **P, resources res){

    projectsh *s,*l;
    int m;

    /*Agenda os projetos obrigatrio na soluo.*/
    schedule_mandatory_projects_in_solution(P,mandatory_projects);

    /*Percorre toda a hash de projetos*/
    HASH_ITER(hh,*projects,s,l){

        /*Gera os meses aleatrios apenas para os projetos no obrigatrio.*/
        if(strcmp(s->project.obrigatory,"N") == 0){

            /*Foi enviesado para que os projetos no terminam nos ltimos meses e
            acabem violando a restrio de pontos de ateno intolerveis.*/
            m = (rand()%60 + s->project.leadTime+1)-15;

            /*Se m > 61, ento retorna 61 seno m.*/
            m = m > 61 ? 61 : m;
            add_solution(P,s->pid,m);
            /*Quando adiciona o novo projeto e a soluo for invlida, ela  deletada da soluo.*/
            if(! (is_solution_valid(P,mandatory_projects,projects,res) )){
                //printf("\n\ns->pid:%d m:%d s->leadtime:%d",s->pid,m,s->project.leadTime);
                delete_from_solution(P,s->pid);
            }
        }
    }

}

solutionhs *ideal_solution(projectsh **projects){

    solutionhs *P = NULL;
    projectsh *s,*l;


    HASH_ITER(hh,*projects,s,l){

        /*Se o projeto no for obrigatrio*/
        if(strcmp(s->project.obrigatory,"N")==0){
            add_solution(&P,s->pid,1);
        }else{
            add_solution(&P,s->pid,s->project.initialMonth);
        }
    }

    return P;
}

