#include "constraints.h"


int consistent_scheduling(solutionhs **P){

    solutionhs *s,*l;

    /*Percorre toda os projetos da solução*/
    HASH_ITER(hh,*P,s,l){

        /*Verifica se está entre os meses 1 e 60*/
        if((s->m < 1) || (s->m > 61)){
            printf("\n\nMes nao esta entre 1 a 61.\n\n");
            return FALSE;
        }
    }

    return TRUE;
}

int check_projects_mandatory_scheduling(projectsh **mandatory_projects, solutionhs **P){

    projectsh *s,*l;
    solutionhs *q;

    HASH_ITER(hh, *mandatory_projects, s,l){

        q = find_solution(*P,s->pid);

        if(q){
            if(q->m != s->project.initialMonth){
                printf("\n\nErro %d! Meses inicais diferentes.\n\n",s->pid);
                return FALSE;
            }
        }
        else{
            printf("\n\nErro %d! Projeto nao encontrado na solucao.\n\n",s->pid);
            return FALSE;
        }
    }
    return TRUE;
}

float yearlycost(solutionhs **P, projectsh **project, char classification[],int year){

    solutionhs *s,*l;
    projectsh *q;
    float soma = 0;

        for(int i = 12 * year-11; i <= 12 * year ; i++){
            /* For each pair (project, month) in the solution*/
            HASH_ITER(hh, *P, s, l){

                /*Não pegar o id que armazena o valor de rp.*/
                if((s->pid != -1) && (s->m < 61)){

                    q = find_project(*project, s->pid);

                    /*Verifica se os recuruso são do mesmo tipo CAPEX ou OPEX*/
                    if(strcmp(classification,q->project.classification) == 0){
                                /*s->m  = mes inicial
                                (s->m + q->project.duration - 1) = mes final
                                i = mes atual*/
                        /*verificar se o mês corrente avaliado está entre o inicial e o final*/

                        if((i >= s->m) && (i <= (s->m + q->project.duration - 1))){
                            //printf("\n\nAno: %d pid %d s->m:%d custo[%d] = %.3f\n",i,q->pid,s->m,i - s->m + 1,q->project.costs[i - s->m]);
                            //printf("Soma atual: %.3f",soma);
                            soma = soma + q->project.costs[i - s->m];

                        }
                    }

                }
            }
        }

    return soma;

}

int check_costs_constraints(solutionhs **P, projectsh **project, resources res){


    float soma_anual;

    for(int y = 1; y < 6; y++){

        soma_anual = yearlycost( P, project,"CAPEX",y);

        //printf("\n\nSoma anual: %.3f e resurso disponivel %.3f",soma_anual,res.capex[y]);

        if(soma_anual > res.capex[y]){
            return FALSE;
        }

    }

    for(int y = 1; y < 6; y++){

        soma_anual = yearlycost( P, project,"OPEX",y);

        //printf("\n\nSoma anual: %.3f e resurso disponivel %.3f",soma_anual,res.opex[y]);

        if(soma_anual > res.opex[y]){
            //printf("\n\nUltrapassou os recursos OPEX disponiveis do ano");
            return FALSE;
        }

    }
    return TRUE;
}

int lead_time(solutionhs **P, projectsh **project){

    solutionhs *s,*l;
    projectsh *q;

    /*Percorre toda a solução*/
    HASH_ITER(hh,*P,s,l){

        /*Não pegar o id -1, ele é usado apenas para armazenar o valor do rp da solução.*/
        if((s->pid != -1) && (s->m < 61)){
            /*Encontra na hash de projetos o projeto = pid, para saber o seu lead_time.*/
            q = find_project(*project,s->pid);

            /*Se o projeto iniciar antes do leadTime retorna falso.*/
            if((s->m) <= q->project.leadTime){
                return FALSE;
            }
        }
    }
    return TRUE;
}



int count_halting_generating_units_per_month(solutionhs **P, projectsh **project, int month, char area[]){


    solutionhs *s,*l;
    projectsh *q;
    int unidade = 0; /*quantidade de unidades geradoras paradas*/

    HASH_ITER(hh, *P, s, l){

        if((s->pid != -1) && (s->m < 61)){

            q  = find_project(*project, s->pid);

            /*verificar se o projeto é de manutenção*/
            if(q->project.maintenance[0] != 'N'){

                /*verificar se o mês corrente avaliado está entre o inicial de parada e o final de parada*/
                if(strcmp(q->project.area,area) == 0){

                    int halting_start = q->project.initialMonth + q->project.halting_start_month;
                    int halting_stop =  halting_start + q->project.halting_duration - 1;

                    if((month >= halting_start) && (month <= halting_stop)){
                        unidade += 1;
                    }
                }
            }
        }
    }


    return unidade;
}

int count_halting_generating_units_maintenance_long(solutionhs **P, projectsh **project,int month, char area[]){

    solutionhs *s,*l;
    projectsh *q;
    int unidade = 0; /*quantidade de unidades geradoras paradas*/

    HASH_ITER(hh, *P, s, l){

        if((s->pid != -1) && (s->m < 61)){

            q = find_project(*project, s->pid);

            /*verificar se o projeto é de manutenção longo*/
            if(q->project.maintenance[0] == 'L'){
                /*verificar se o mês corrente avaliado está entre o inicial de parada e o final de parada*/
                if(strcmp(q->project.area,area) == 0){

                    int halting_start = q->project.initialMonth + q->project.halting_start_month;
                    int halting_stop =  halting_start + q->project.halting_duration - 1;

                    if((month >= halting_start) && (month <= halting_stop)){
                        unidade += 1;
                    }
                }
            }
        }
    }
    return unidade;
}

int RP_Location(solutionhs **P, projectsh **project){

    location vet[3];  /*Recebe CAC, EUC e LMO*/
    int u1,u2;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    strcpy(vet[0].usina,"CAC");
    strcpy(vet[1].usina,"EUC");
    strcpy(vet[2].usina,"LMO");

    for(int m = 1; m <= 2*T; m++){

        for(int i = 0; i < 3; i++){
            for(int j = i+1; j < 3; j++){

               u1 = count_halting_generating_units_per_month(P,project,m,vet[i].usina);
               u2 = count_halting_generating_units_per_month(P,project,m,vet[j].usina);

               /*printf("\nUsina: %s Mes: %d u1: %d",vet[i].usina,m,u1);
               printf("\nUsina: %s Mes: %d u2: %d",vet[j].usina,m,u2);*/

               if((u1 >= 2) && (u2 != 0)){
               //     printf("\n\nu1 = %d u2 = %d",u1,u2);
               //     printf("\n\nErro na restricao RP location");
                    return FALSE;

               }

               u1 = count_halting_generating_units_per_month(P,project,m,vet[j].usina);
               u2 = count_halting_generating_units_per_month(P,project,m,vet[i].usina);

               /*printf("\nUsina: %s Mes: %d u1: %d",vet[j].usina,m,u1);
               printf("\nUsina: %s Mes: %d u2: %d",vet[i].usina,m,u2);*/

               if((u1 >= 2) && (u2 != 0)){
               //     printf("\n\nu1 = %d u2 = %d",u1,u2);
               //     printf("\n\nErro na restricao RP location");
                    return FALSE;


               }
            }
        }
    }
    return TRUE;
}

int BBB_Location1(solutionhs **P, projectsh **project){

    int u1,u2;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    for(int m = 1; m <= 2*T; m++){

        u1 = count_halting_generating_units_per_month(P, project, m, "BAB");
        u2 = count_halting_generating_units_per_month(P, project, m, "BAR");

        /*printf("\nUsina: BAB Mes: %d u1: %d",m,u1);
        printf("\nUsina: BAR Mes: %d u2: %d",m,u2);*/

        if((u1 >= 2) && (u2 != 0)){
       //     printf("\n\nm = %d u1 = %d u2 = %d",m,u1,u2);
       //     printf("\n\nErro na restrição BBB1 location");
            return FALSE;
        }

        u1 = count_halting_generating_units_per_month(P,project,m,"BAR");
        u2 = count_halting_generating_units_per_month(P,project,m,"BAB");

       /* printf("\nUsina: BAR Mes: %d u1: %d",m,u1);
        printf("\nUsina: BAB Mes: %d u2: %d",m,u2);*/

        if((u1 >= 2) && (u2 != 0)){
          //  printf("\n\nm = %d u1 = %d u2 = %d",m,u1,u2);
       //     printf("\n\nErro na restrição BBB1 location");
            return FALSE;
        }

    }

    return TRUE;
}

int BBB_Location2(solutionhs **P, projectsh **project){

    int u1;

    for(int m = 0; m < 2*T; m++){

        u1 = count_halting_generating_units_per_month(P,project,m,"BAR");

        if(u1 > 2){
           // printf("\n\nm = %d u1 = %d",m,u1);
          //  printf("\n\nErro na restrição BBB2 location");
            return FALSE;
        }
    }

    return TRUE;

}

int BBB_Location3(solutionhs **P, projectsh **project){


    int u1,u2;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    for(int m = 1; m <= 2*T; m++){

        u1 = count_halting_generating_units_per_month(P,project,m,"BAB");
        u2 = count_halting_generating_units_per_month(P,project,m,"BAR");

       /* printf("\nUsina: BAB Mes: %d u1: %d",m,u1);
        printf("\nUsina: BAR Mes: %d u2: %d",m,u2);*/

        if((u1 == 2) && (u2 != 0)){
         //   printf("\n\nm = %d u1 = %d u2 = %d",m,u1,u2);
         //-   printf("\n\nErro na restrição BBB3 location");
            return FALSE;

        }
    }

    return TRUE;
}

int NAPI_Location(solutionhs **P, projectsh **project){

    int u1,u2,u3;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    for(int m = 1; m <= 2*T; m++){

        u1 = count_halting_generating_units_per_month(P,project,m,"NAV");
        u2 = count_halting_generating_units_per_month(P,project,m,"PRO");
        u3 = count_halting_generating_units_per_month(P,project,m,"IBI");

        /*printf("\n\nUsina: NAV Mes: %d u1: %d",m,u1);
        printf("\nUsina: PRO Mes: %d u2: %d",m,u2);
        printf("\nUsina: IBI Mes: %d u3: %d",m,u3);*/

        if((u1 + u2 + u3) > 2){
         //   printf("\n\nu1 = %d u2 = %d u3 = %d ",u1,u2,u3);
         //   printf("\n\nErro na restricao NAPI location");
            return FALSE;

        }

    }
    return TRUE;

}

int AV_Location(solutionhs **P, projectsh **project){

    int u1;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    for(int m = 1; m <= 2*T; m++){

        u1 = count_halting_generating_units_maintenance_long(P,project,m,"AGV");

        if(u1 > 2){
          //  printf("\n\nu1 = %d",u1);
          //  printf("\n\nErro na restricao AV location");
            return FALSE;

        }

    }
    return TRUE;

}

int Halting_restrictions_AGV_plant(solutionhs **P, projectsh **project){

    int u1,u2,u3,u4;  /*Recebe a quantidade de usinas parada em determinado mes para determina usina*/

    for(int m = 1; m <= 30; m++){

        u1 = count_halting_generating_units_per_month(P,project,m,"AGV");
        u2 = count_halting_generating_units_per_month(P,project,m,"BAB");
        u3 = count_halting_generating_units_per_month(P,project,m,"NAV");
        u4 = count_halting_generating_units_per_month(P,project,m,"PRO");

        /*printf("\n\nUsina: AGV Mes: %d u1: %d",m,u1);
        printf("\n\nUsina: BAB Mes: %d u1: %d",m,u2);
        printf("\n\nUsina: NAV Mes: %d u1: %d",m,u3);
        printf("\n\nUsina: PRO Mes: %d u1: %d",m,u4);*/

        if((u1 >= 2) && ((u2 != 0) || (u3 != 0) || (u4 != 0))){
          //  printf("\n\nu1 = %d u2 = %d u3 = %d u4 = %d",u1,u2,u3,u4);
          //  printf("\n\nErro na restricao halting restrictions AGV plant");
            return FALSE;
        }

    }
    return TRUE;
}

int Halting_restrictions_division(solutionhs **P, projectsh **project){

    int u1,u2,u3,u4,u5,u6; /*Recebe a quantidade de unidades geradoras paradas.*/

    for(int m = 1 ; m <= 2*T; m++){
    /*Divisão: COG Usina: COG*/
        u1 = count_halting_generating_units_per_month(P,project,m,"COG");
        if(u1 > 3){
          //  printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao COG");
         //   printf("\nu1: %d",u1);
            return FALSE;
        }

        /*Divisão: IT Usina: IT*/
        u1 = count_halting_generating_units_per_month(P,project,m,"IT");
        if(u1 > 3){
           // printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao IT");
          //  printf("\nu1: %d",u1);
            return FALSE;

        }

        /*Divisão: AGV Usina: AGV*/
        u1 = count_halting_generating_units_per_month(P,project,m,"AGV");
        if(u1 > 3){
           // printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao AGV");
         //   printf("\nu1: %d",u1);
            return FALSE;

        }

        /*Divisão: PRO Usinas: NAV e PRO*/
        u1 = count_halting_generating_units_per_month(P,project,m,"NAV");
        u2 = count_halting_generating_units_per_month(P,project,m,"PRO");
        if(u1 + u2 > 3){

           // printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao PRO");
          //  printf("\nu1: %d u2: %d",u1,u2);
            return FALSE;

        }

        /*Divisão: BAR Usinas: BAB, BAR, IBI*/
        u1 = count_halting_generating_units_per_month(P,project,m,"BAB");
        u2 = count_halting_generating_units_per_month(P,project,m,"BAR");
        u3 = count_halting_generating_units_per_month(P,project,m,"IBI");

        if(u1 + u2 + u3 > 3){
            //printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao BAR");
           // printf("\nu1: %d u2: %d u3: %d",u1,u2,u3);
            return FALSE;

        }

        /*Divisão LMO Usinas: CAC, EUC, LMO, MOG, SJS, SJQ*/
        u1 = count_halting_generating_units_per_month(P,project,m,"CAC");
        u2 = count_halting_generating_units_per_month(P,project,m,"EUC");
        u3 = count_halting_generating_units_per_month(P,project,m,"LMO");
        u4 = count_halting_generating_units_per_month(P,project,m,"MOG");
        u5 = count_halting_generating_units_per_month(P,project,m,"SJS");
        u6 = count_halting_generating_units_per_month(P,project,m,"SJQ");

        if(u1 + u2 + u3 + u4 + u5 + u6 > 3){
           // printf("\n\nErro, mais de 3 unidades geradoras paradas na divisao LMO");
            //printf("\nu1: %d u2: %d u3: %d u4: %d u5: %d u6: %d",u1,u2,u3,u4,u5,u6);
            return FALSE;

        }
    }

    return TRUE;
}

int is_solution_valid(solutionhs **P,projectsh **mandatory_projects,projectsh **project,resources res){


    /*if( consistent_scheduling(P)){
        printf("\n\nErro no consistent_scheduling");
        return FALSE;
    }*/

    if(!(check_projects_mandatory_scheduling(mandatory_projects,P))){  //Verifica se os projetos obrigatoris estão na solução
      //  printf("\n\nErro no projetos mandatory");
        return FALSE;
    }

    if(!(check_costs_constraints(P,project,res))){  /*Verifica se o custos anuais não ultrapassam os recursos*/
        //printf("\n\nErro check_costs_contrains");
        return FALSE;
    }

    if(!(lead_time(P,project))){
       // printf("\n\nErro no leadtime");
        return FALSE;
    }


    if(!(RP_Location(P,project))){              /*Não pode possuir duas ou mais usinas paradas em RP*/
      // printf("\n\nErro RP_Location");
       return FALSE;
    }

    if(!(BBB_Location1(P,project))){
      //  printf("\n\nErro BBB_Location1");
        return FALSE;
    }

    if(!(BBB_Location2(P,project))){
       // printf("\n\nErro BBB_Location2");
        return FALSE;
    }

    if(!( BBB_Location3(P,project))){
      //  printf("\n\nErro BBB_Location3");
        return FALSE;
    }
    if(!( NAPI_Location(P,project))){
      //  printf("\n\nErro NAPI_location");
        return FALSE;
    }
    if(!( AV_Location(P,project))){
      //  printf("\n\nErro AV_Location");
        return FALSE;

    }

    if(!( Halting_restrictions_AGV_plant(P,project))){
      //  printf("\n\nErro Halting_restrictions_AGV_plant");
        return FALSE;
    }

    if(!( Halting_restrictions_division(P,project))){
        //printf("\n\nErro Halting_restrictions_division");
        return FALSE;
    }

    return TRUE;
}
int intolerable_points(warning_pointshs **wp,projectsh **project,solutionhs **P){

    warning_pointshs *s,*l;
    projectsh *t;
    solutionhs *q;
    int max;

    /*Percorre toda a hash de pontos de atenção.*/
    HASH_ITER(hh,*wp,s,l){

        /*Incializa o max a cada iteração.*/
        max = 0;
        /*Pega apenas os pontos de atenção intoleráveis.*/
        if(strcmp(s->warning_points.risk_category,"I") == 0){

            /*Percorre o grupo de projetos que controlam esse ponto de atenção.*/
            for(int i = 0; i < s->warning_points.tam; i++){

                /*Encontra o projeto na hash de solução.*/
                if(!(q = find_solution(*P,s->warning_points.group_of_projects[i]) ) ){
                    continue;
                }
                
                /*Encontra o projeto na hash de projeto para acessar os atributos.*/
                t = find_project(*project,s->warning_points.group_of_projects[i]);

                //printf("mes inicial: %d duracao: %d\n",q->m,t->project.duration);
                /*Se max for menor que o mês agendado na solução + duração entao max recebe essa soma.*/
                if(q->m + t->project.duration - 1 > max){

                    max = q->m + t->project.duration-1;
                    //printf("\n\nmax:%d ",max);

                }
            }

        }


       /*Se o valor max for maior que o deadline então retorna falso.*/
        if((s->warning_points.deadline != -1)&&(max > s->warning_points.deadline)){
            //printf("\n\ns->warning_points.id: %d max:%d deadline:%d",s->pid , max, s->warning_points.deadline);
            return FALSE;
        }
    }
    return TRUE;
}